@@ -366,7 +366,6 @@ func (l *localPortWatcher) SyncServices(serviceInterface []interface{}) {
 	}
 	for _, chain := range []string{iptableNodePortChain, iptableExternalIPChain} {
 		recreateIPTRules("nat", chain, keepIPTRules)
-		recreateIPTRules("filter", chain, keepIPTRules)
 	}
 	removeStaleRoutes(keepRoutes)
 }
@@ -421,15 +420,6 @@ func getLoadBalancerIPTRules(svc *kapi.Service, svcPort kapi.ServicePort, gatewa
 				"-j", "DNAT", "--to-destination", util.JoinHostPortInt32(gatewayIP, targetPort),
 			},
 		})
-		rules = append(rules, iptRule{
-			table: "filter",
-			chain: iptableNodePortChain,
-			args: []string{
-				"-d", ing.IP,
-				"-p", string(svcPort.Protocol), "--dport", ingPort,
-				"-j", "ACCEPT",
-			},
-		})
 	}
 	return rules
 }
